/*
 * DO NOT EDIT.  THIS FILE IS GENERATED FROM /builds/slave/mozilla-central-linux64-xulrunner/build/dom/interfaces/html/nsIDOMNSHTMLLabelElement.idl
 */

#ifndef __gen_nsIDOMNSHTMLLabelElement_h__
#define __gen_nsIDOMNSHTMLLabelElement_h__


#ifndef __gen_domstubs_h__
#include "domstubs.h"
#endif

/* For IDL files that don't want to include root IDL files. */
#ifndef NS_NO_VTABLE
#define NS_NO_VTABLE
#endif
class nsIDOMHTMLElement; /* forward declaration */


/* starting interface:    nsIDOMNSHTMLLabelElement */
#define NS_IDOMNSHTMLLABELELEMENT_IID_STR "93e8a513-6cef-44c3-875e-1bce83c3b1f0"

#define NS_IDOMNSHTMLLABELELEMENT_IID \
  {0x93e8a513, 0x6cef, 0x44c3, \
    { 0x87, 0x5e, 0x1b, 0xce, 0x83, 0xc3, 0xb1, 0xf0 }}

class NS_NO_VTABLE NS_SCRIPTABLE nsIDOMNSHTMLLabelElement : public nsISupports {
 public: 

  NS_DECLARE_STATIC_IID_ACCESSOR(NS_IDOMNSHTMLLABELELEMENT_IID)

  /* readonly attribute nsIDOMHTMLElement control; */
  NS_SCRIPTABLE NS_IMETHOD GetControl(nsIDOMHTMLElement * *aControl) = 0;

};

  NS_DEFINE_STATIC_IID_ACCESSOR(nsIDOMNSHTMLLabelElement, NS_IDOMNSHTMLLABELELEMENT_IID)

/* Use this macro when declaring classes that implement this interface. */
#define NS_DECL_NSIDOMNSHTMLLABELELEMENT \
  NS_SCRIPTABLE NS_IMETHOD GetControl(nsIDOMHTMLElement * *aControl); 

/* Use this macro to declare functions that forward the behavior of this interface to another object. */
#define NS_FORWARD_NSIDOMNSHTMLLABELELEMENT(_to) \
  NS_SCRIPTABLE NS_IMETHOD GetControl(nsIDOMHTMLElement * *aControl) { return _to GetControl(aControl); } 

/* Use this macro to declare functions that forward the behavior of this interface to another object in a safe way. */
#define NS_FORWARD_SAFE_NSIDOMNSHTMLLABELELEMENT(_to) \
  NS_SCRIPTABLE NS_IMETHOD GetControl(nsIDOMHTMLElement * *aControl) { return !_to ? NS_ERROR_NULL_POINTER : _to->GetControl(aControl); } 

#if 0
/* Use the code below as a template for the implementation class for this interface. */

/* Header file */
class nsDOMNSHTMLLabelElement : public nsIDOMNSHTMLLabelElement
{
public:
  NS_DECL_ISUPPORTS
  NS_DECL_NSIDOMNSHTMLLABELELEMENT

  nsDOMNSHTMLLabelElement();

private:
  ~nsDOMNSHTMLLabelElement();

protected:
  /* additional members */
};

/* Implementation file */
NS_IMPL_ISUPPORTS1(nsDOMNSHTMLLabelElement, nsIDOMNSHTMLLabelElement)

nsDOMNSHTMLLabelElement::nsDOMNSHTMLLabelElement()
{
  /* member initializers and constructor code */
}

nsDOMNSHTMLLabelElement::~nsDOMNSHTMLLabelElement()
{
  /* destructor code */
}

/* readonly attribute nsIDOMHTMLElement control; */
NS_IMETHODIMP nsDOMNSHTMLLabelElement::GetControl(nsIDOMHTMLElement * *aControl)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

/* End of implementation class template. */
#endif


#endif /* __gen_nsIDOMNSHTMLLabelElement_h__ */
